@@ -759,6 +759,20 @@ seat_reconfigure(struct server *server)
 	}
 }
 
+void
+seat_force_focus_surface(struct seat *seat, struct wlr_surface *surface)
+{
+	if (seat->server->session_lock_manager->locked) {
+		return;
+	}
+	uint32_t *pressed_sent_keycodes = key_state_pressed_sent_keycodes();
+	int nr_pressed_sent_keycodes = key_state_nr_pressed_sent_keycodes();
+	struct wlr_keyboard *kb = &seat->keyboard_group->keyboard;
+
+	wlr_seat_keyboard_enter(seat->seat, surface,
+		pressed_sent_keycodes, nr_pressed_sent_keycodes, &kb->modifiers);
+}
+
 static void
 seat_focus(struct seat *seat, struct wlr_surface *surface,
 		bool replace_exclusive_layer, bool is_lock_surface)
@@ -902,12 +916,12 @@ seat_focus_override_begin(struct seat *seat, enum input_mode input_mode,
 }
 
 void
-seat_focus_override_end(struct seat *seat)
+seat_focus_override_end(struct seat *seat, bool restore_focus)
 {
 	seat->server->input_mode = LAB_INPUT_STATE_PASSTHROUGH;
 
 	if (seat->focus_override.surface) {
-		if (!seat->seat->keyboard_state.focused_surface) {
+		if (restore_focus) {
 			seat_focus(seat, seat->focus_override.surface,
 				/*replace_exclusive_layer*/ false,
 				/*is_lock_surface*/ false);
@@ -916,5 +930,7 @@ seat_focus_override_end(struct seat *seat)
 		seat->focus_override.surface = NULL;
 	}
 
-	cursor_update_focus(seat->server);
+	if (restore_focus) {
+		cursor_update_focus(seat->server);
+	}
 }
