@@ -1,5 +1,6 @@
 // SPDX-License-Identifier: GPL-2.0-only
 #define _POSIX_C_SOURCE 200809L
+#include "workspaces.h"
 #include <assert.h>
 #include <cairo.h>
 #include <pango/pangocairo.h>
@@ -7,19 +8,21 @@
 #include <stdlib.h>
 #include <string.h>
 #include <strings.h>
-#include "config.h"
+#include <wlr/types/wlr_output_layout.h>
+#include <wlr/types/wlr_scene.h>
 #include "buffer.h"
 #include "common/font.h"
 #include "common/graphic-helpers.h"
 #include "common/list.h"
 #include "common/mem.h"
+#include "config/rcxml.h"
 #include "input/keyboard.h"
 #include "labwc.h"
+#include "output.h"
 #include "protocols/cosmic-workspaces.h"
 #include "protocols/ext-workspace.h"
+#include "theme.h"
 #include "view.h"
-#include "workspaces.h"
-#include "xwayland.h"
 
 #define COSMIC_WORKSPACES_VERSION 1
 #define EXT_WORKSPACES_VERSION 1
@@ -167,12 +170,8 @@ _osd_update(struct server *server)
 		struct wlr_box output_box;
 		wlr_output_layout_get_box(output->server->output_layout,
 			output->wlr_output, &output_box);
-		int lx = output->usable_area.x
-			+ (output->usable_area.width - width) / 2
-			+ output_box.x;
-		int ly = output->usable_area.y
-			+ (output->usable_area.height - height) / 2
-			+ output_box.y;
+		int lx = output_box.x + (output_box.width - width) / 2;
+		int ly = output_box.y + (output_box.height - height) / 2;
 		wlr_scene_node_set_position(&output->workspace_osd->node, lx, ly);
 		wlr_scene_buffer_set_buffer(output->workspace_osd, &buffer->base);
 		wlr_scene_buffer_set_dest_size(output->workspace_osd,
@@ -183,6 +182,33 @@ _osd_update(struct server *server)
 	}
 }
 
+static struct workspace *
+workspace_find_by_name(struct server *server, const char *name)
+{
+	struct workspace *workspace;
+
+	/* by index */
+	size_t parsed_index = parse_workspace_index(name);
+	if (parsed_index) {
+		size_t index = 0;
+		wl_list_for_each(workspace, &server->workspaces.all, link) {
+			if (parsed_index == ++index) {
+				return workspace;
+			}
+		}
+	}
+
+	/* by name */
+	wl_list_for_each(workspace, &server->workspaces.all, link) {
+		if (!strcmp(workspace->name, name)) {
+			return workspace;
+		}
+	}
+
+	wlr_log(WLR_ERROR, "Workspace '%s' not found", name);
+	return NULL;
+}
+
 /* cosmic workspace handlers */
 static void
 handle_cosmic_workspace_activate(struct wl_listener *listener, void *data)
@@ -208,20 +234,19 @@ add_workspace(struct server *server, const char *name)
 	struct workspace *workspace = znew(*workspace);
 	workspace->server = server;
 	workspace->name = xstrdup(name);
-	workspace->tree = wlr_scene_tree_create(server->view_tree);
+	workspace->tree = wlr_scene_tree_create(server->workspace_tree);
+	workspace->view_trees[VIEW_LAYER_ALWAYS_ON_BOTTOM] =
+		wlr_scene_tree_create(workspace->tree);
+	workspace->view_trees[VIEW_LAYER_NORMAL] =
+		wlr_scene_tree_create(workspace->tree);
+	workspace->view_trees[VIEW_LAYER_ALWAYS_ON_TOP] =
+		wlr_scene_tree_create(workspace->tree);
 	wl_list_append(&server->workspaces.all, &workspace->link);
-	if (!server->workspaces.current) {
-		server->workspaces.current = workspace;
-	} else {
-		wlr_scene_node_set_enabled(&workspace->tree->node, false);
-	}
-
-	bool active = server->workspaces.current == workspace;
+	wlr_scene_node_set_enabled(&workspace->tree->node, false);
 
 	/* cosmic */
 	workspace->cosmic_workspace = lab_cosmic_workspace_create(server->workspaces.cosmic_group);
 	lab_cosmic_workspace_set_name(workspace->cosmic_workspace, name);
-	lab_cosmic_workspace_set_active(workspace->cosmic_workspace, active);
 
 	workspace->on_cosmic.activate.notify = handle_cosmic_workspace_activate;
 	wl_signal_add(&workspace->cosmic_workspace->events.activate,
@@ -232,7 +257,6 @@ add_workspace(struct server *server, const char *name)
 		server->workspaces.ext_manager, /*id*/ NULL);
 	lab_ext_workspace_assign_to_group(workspace->ext_workspace, server->workspaces.ext_group);
 	lab_ext_workspace_set_name(workspace->ext_workspace, name);
-	lab_ext_workspace_set_active(workspace->ext_workspace, active);
 
 	workspace->on_ext.activate.notify = handle_ext_workspace_activate;
 	wl_signal_add(&workspace->ext_workspace->events.activate,
@@ -395,10 +419,31 @@ workspaces_init(struct server *server)
 
 	wl_list_init(&server->workspaces.all);
 
-	struct workspace *conf;
+	struct workspace_config *conf;
 	wl_list_for_each(conf, &rc.workspace_config.workspaces, link) {
 		add_workspace(server, conf->name);
 	}
+
+	/*
+	 * After adding workspaces, check if there is an initial workspace
+	 * selected and set that as the initial workspace.
+	 */
+	char *initial_name = rc.workspace_config.initial_workspace_name;
+	struct workspace *initial = NULL;
+	struct workspace *first = wl_container_of(
+		server->workspaces.all.next, first, link);
+
+	if (initial_name) {
+		initial = workspace_find_by_name(server, initial_name);
+	}
+	if (!initial) {
+		initial = first;
+	}
+
+	server->workspaces.current = initial;
+	wlr_scene_node_set_enabled(&initial->tree->node, true);
+	lab_cosmic_workspace_set_active(initial->cosmic_workspace, true);
+	lab_ext_workspace_set_active(initial->ext_workspace, true);
 }
 
 /*
@@ -424,11 +469,13 @@ workspaces_switch_to(struct workspace *target, bool update_focus)
 	lab_ext_workspace_set_active(
 		server->workspaces.current->ext_workspace, false);
 
-	/* Move Omnipresent views to new workspace */
+	/*
+	 * Move Omnipresent views to new workspace.
+	 * Not using for_each_view() since it skips views that
+	 * view_is_focusable() returns false (e.g. Conky).
+	 */
 	struct view *view;
-	enum lab_view_criteria criteria =
-		LAB_VIEW_CRITERIA_CURRENT_WORKSPACE;
-	for_each_view_reverse(view, &server->views, criteria) {
+	wl_list_for_each_reverse(view, &server->views, link) {
 		if (view->visible_on_all_workspaces) {
 			view_move_to_workspace(view, target);
 		}
@@ -444,24 +491,17 @@ workspaces_switch_to(struct workspace *target, bool update_focus)
 	server->workspaces.current = target;
 
 	struct view *grabbed_view = server->grabbed_view;
-	if (grabbed_view && !view_is_always_on_top(grabbed_view)) {
+	if (grabbed_view) {
 		view_move_to_workspace(grabbed_view, target);
 	}
 
 	/*
 	 * Make sure we are focusing what the user sees. Only refocus if
-	 * the focus is not already on an omnipresent or always-on-top view.
-	 *
-	 * TODO: Decouple always-on-top views from the omnipresent state.
-	 *       One option for that would be to create a new scene tree
-	 *       as child of every workspace tree and then reparent a-o-t
-	 *       windows to that one. Combined with adjusting the condition
-	 *       below that should take care of the issue.
+	 * the focus is not already on an omnipresent view.
 	 */
 	if (update_focus) {
 		struct view *active_view = server->active_view;
-		if (!active_view || (!active_view->visible_on_all_workspaces
-				&& !view_is_always_on_top(active_view))) {
+		if (!(active_view && active_view->visible_on_all_workspaces)) {
 			desktop_focus_topmost_view(server);
 		}
 	}
@@ -508,21 +548,13 @@ workspaces_find(struct workspace *anchor, const char *name, bool wrap)
 	if (!name) {
 		return NULL;
 	}
-	size_t index = 0;
-	struct workspace *target;
-	size_t wants_index = parse_workspace_index(name);
-	struct wl_list *workspaces = &anchor->server->workspaces.all;
+	struct server *server = anchor->server;
+	struct wl_list *workspaces = &server->workspaces.all;
 
-	if (wants_index) {
-		wl_list_for_each(target, workspaces, link) {
-			if (wants_index == ++index) {
-				return target;
-			}
-		}
-	} else if (!strcasecmp(name, "current")) {
+	if (!strcasecmp(name, "current")) {
 		return anchor;
 	} else if (!strcasecmp(name, "last")) {
-		return anchor->server->workspaces.last;
+		return server->workspaces.last;
 	} else if (!strcasecmp(name, "left")) {
 		return get_prev(anchor, workspaces, wrap);
 	} else if (!strcasecmp(name, "right")) {
@@ -531,15 +563,8 @@ workspaces_find(struct workspace *anchor, const char *name, bool wrap)
 		return get_prev_occupied(anchor, workspaces, wrap);
 	} else if (!strcasecmp(name, "right-occupied")) {
 		return get_next_occupied(anchor, workspaces, wrap);
-	} else {
-		wl_list_for_each(target, workspaces, link) {
-			if (!strcasecmp(target->name, name)) {
-				return target;
-			}
-		}
 	}
-	wlr_log(WLR_ERROR, "Workspace '%s' not found", name);
-	return NULL;
+	return workspace_find_by_name(server, name);
 }
 
 static void
@@ -566,68 +591,66 @@ workspaces_reconfigure(struct server *server)
 	 *   - Destroy workspaces if fewer workspace are desired
 	 */
 
-	struct wl_list *actual_workspace_link = server->workspaces.all.next;
+	struct wl_list *workspace_link = server->workspaces.all.next;
 
-	struct workspace *configured_workspace;
-	wl_list_for_each(configured_workspace,
-			&rc.workspace_config.workspaces, link) {
-		struct workspace *actual_workspace = wl_container_of(
-			actual_workspace_link, actual_workspace, link);
+	struct workspace_config *conf;
+	wl_list_for_each(conf, &rc.workspace_config.workspaces, link) {
+		struct workspace *workspace = wl_container_of(
+			workspace_link, workspace, link);
 
-		if (actual_workspace_link == &server->workspaces.all) {
+		if (workspace_link == &server->workspaces.all) {
 			/* # of configured workspaces increased */
 			wlr_log(WLR_DEBUG, "Adding workspace \"%s\"",
-				configured_workspace->name);
-			add_workspace(server, configured_workspace->name);
+				conf->name);
+			add_workspace(server, conf->name);
 			continue;
 		}
-		if (strcmp(actual_workspace->name, configured_workspace->name)) {
+		if (strcmp(workspace->name, conf->name)) {
 			/* Workspace is renamed */
 			wlr_log(WLR_DEBUG, "Renaming workspace \"%s\" to \"%s\"",
-				actual_workspace->name, configured_workspace->name);
-			free(actual_workspace->name);
-			actual_workspace->name = xstrdup(configured_workspace->name);
+				workspace->name, conf->name);
+			xstrdup_replace(workspace->name, conf->name);
 			lab_cosmic_workspace_set_name(
-				actual_workspace->cosmic_workspace, actual_workspace->name);
+				workspace->cosmic_workspace, workspace->name);
 			lab_ext_workspace_set_name(
-				actual_workspace->ext_workspace, actual_workspace->name);
+				workspace->ext_workspace, workspace->name);
 		}
-		actual_workspace_link = actual_workspace_link->next;
+		workspace_link = workspace_link->next;
 	}
 
-	if (actual_workspace_link == &server->workspaces.all) {
+	if (workspace_link == &server->workspaces.all) {
 		return;
 	}
 
 	/* # of configured workspaces decreased */
-	overlay_hide(&server->seat);
+	overlay_finish(&server->seat);
 	struct workspace *first_workspace =
 		wl_container_of(server->workspaces.all.next, first_workspace, link);
 
-	while (actual_workspace_link != &server->workspaces.all) {
-		struct workspace *actual_workspace = wl_container_of(
-			actual_workspace_link, actual_workspace, link);
+	while (workspace_link != &server->workspaces.all) {
+		struct workspace *workspace = wl_container_of(
+			workspace_link, workspace, link);
 
 		wlr_log(WLR_DEBUG, "Destroying workspace \"%s\"",
-			actual_workspace->name);
+			workspace->name);
 
 		struct view *view;
 		wl_list_for_each(view, &server->views, link) {
-			if (view->workspace == actual_workspace) {
+			if (view->workspace == workspace) {
 				view_move_to_workspace(view, first_workspace);
 			}
 		}
 
-		if (server->workspaces.current == actual_workspace) {
+		if (server->workspaces.current == workspace) {
 			workspaces_switch_to(first_workspace,
 				/* update_focus */ true);
 		}
-		if (server->workspaces.last == actual_workspace) {
+		if (server->workspaces.last == workspace) {
 			server->workspaces.last = first_workspace;
 		}
 
-		actual_workspace_link = actual_workspace_link->next;
-		destroy_workspace(actual_workspace);
+		workspace_link = workspace_link->next;
+		destroy_workspace(workspace);
 	}
 }
 
