@@ -225,8 +225,7 @@ handle_commit(struct wl_listener *listener, void *data)
 			set_fullscreen_from_request(view, &toplevel->requested);
 		}
 		if (toplevel->requested.maximized) {
-			view_maximize(view, VIEW_AXIS_BOTH,
-				/*store_natural_geometry*/ true);
+			view_maximize(view, VIEW_AXIS_BOTH);
 		}
 		return;
 	}
@@ -455,11 +454,11 @@ handle_request_move(struct wl_listener *listener, void *data)
 	 * the provided serial against a list of button press serials sent to
 	 * this client, to prevent the client from requesting this whenever they
 	 * want.
+	 *
+	 * Note: interactive_begin() checks that view == server->grabbed_view.
 	 */
 	struct view *view = wl_container_of(listener, view, request_move);
-	if (view == view->server->seat.pressed.ctx.view) {
-		interactive_begin(view, LAB_INPUT_STATE_MOVE, LAB_EDGE_NONE);
-	}
+	interactive_begin(view, LAB_INPUT_STATE_MOVE, LAB_EDGE_NONE);
 }
 
 static void
@@ -472,12 +471,12 @@ handle_request_resize(struct wl_listener *listener, void *data)
 	 * the provided serial against a list of button press serials sent to
 	 * this client, to prevent the client from requesting this whenever they
 	 * want.
+	 *
+	 * Note: interactive_begin() checks that view == server->grabbed_view.
 	 */
 	struct wlr_xdg_toplevel_resize_event *event = data;
 	struct view *view = wl_container_of(listener, view, request_resize);
-	if (view == view->server->seat.pressed.ctx.view) {
-		interactive_begin(view, LAB_INPUT_STATE_RESIZE, event->edges);
-	}
+	interactive_begin(view, LAB_INPUT_STATE_RESIZE, event->edges);
 }
 
 static void
@@ -501,12 +500,11 @@ handle_request_maximize(struct wl_listener *listener, void *data)
 		return;
 	}
 
-	if (!view->mapped && !view->output) {
+	if (!view->mapped && !output_is_usable(view->output)) {
 		view_set_output(view, output_nearest_to_cursor(view->server));
 	}
 	bool maximized = toplevel->requested.maximized;
-	view_maximize(view, maximized ? VIEW_AXIS_BOTH : VIEW_AXIS_NONE,
-		/*store_natural_geometry*/ true);
+	view_maximize(view, maximized ? VIEW_AXIS_BOTH : VIEW_AXIS_NONE);
 }
 
 static void
@@ -523,7 +521,7 @@ handle_request_fullscreen(struct wl_listener *listener, void *data)
 		return;
 	}
 
-	if (!view->mapped && !view->output) {
+	if (!view->mapped && !output_is_usable(view->output)) {
 		view_set_output(view, output_nearest_to_cursor(view->server));
 	}
 	set_fullscreen_from_request(view,
@@ -821,7 +819,7 @@ handle_map(struct wl_listener *listener, void *data)
 	 * An output should have been chosen when the surface was first
 	 * created, but take one more opportunity to assign an output if not.
 	 */
-	if (!view->output) {
+	if (!output_is_usable(view->output)) {
 		view_set_output(view, output_nearest_to_cursor(view->server));
 	}
 
@@ -1021,7 +1019,8 @@ handle_new_xdg_toplevel(struct wl_listener *listener, void *data)
 	}
 
 	view->workspace = server->workspaces.current;
-	view->scene_tree = wlr_scene_tree_create(view->workspace->tree);
+	view->scene_tree = wlr_scene_tree_create(
+		view->workspace->view_trees[VIEW_LAYER_NORMAL]);
 	wlr_scene_node_set_enabled(&view->scene_tree->node, false);
 
 	struct wlr_scene_tree *tree = wlr_scene_xdg_surface_create(
